@@ -24,10 +24,11 @@
 #undef RES_C
 
 #ifndef lint
-static  char rcsid[] = "@(#)$Id: res.c,v 1.21.2.15 2003/10/12 22:56:33 chopin Exp $";
+static const volatile char rcsid[] = "@(#)$Id: res.c,v 1.42 2004/11/02 16:32:01 chopin Exp $";
 #endif
 
-/* #undef	DEBUG	/* because there is a lot of debug code in here :-) */
+/* because there is a lot of debug code in here :-) */
+/* #undef	DEBUG */
 
 static	char	hostbuf[HOSTLEN+1+100]; /* +100 for INET6 */
 static	char	dot[] = ".";
@@ -36,24 +37,24 @@ static	CacheTable	hashtable[ARES_CACSIZE];
 static	aCache	*cachetop = NULL;
 static	ResRQ	*last, *first;
 
-static	void	rem_cache __P((aCache *));
-static	void	rem_request __P((ResRQ *));
-static	int	do_query_name __P((Link *, char *, ResRQ *, int));
-static	int	do_query_number __P((Link *, struct IN_ADDR *, ResRQ *));
-static	void	resend_query __P((ResRQ *));
-static	int	proc_answer __P((ResRQ *, HEADER *, char *, char *));
-static	int	query_name __P((char *, int, int, ResRQ *));
-static	aCache	*make_cache __P((ResRQ *)), *rem_list __P((aCache *));
-static	aCache	*find_cache_name __P((char *));
-static	aCache	*find_cache_number __P((ResRQ *, char *));
-static	int	add_request __P((ResRQ *));
-static	ResRQ	*make_request __P((Link *));
-static	int	send_res_msg __P((char *, int, int));
-static	ResRQ	*find_id __P((int));
-static	int	hash_number __P((unsigned char *));
-static	void	update_list __P((ResRQ *, aCache *));
-static	int	hash_name __P((char *));
-static	int	bad_hostname __P((char *, int));
+static	void	rem_cache (aCache *);
+static	void	rem_request (ResRQ *);
+static	int	do_query_name (Link *, char *, ResRQ *, int);
+static	int	do_query_number (Link *, struct IN_ADDR *, ResRQ *);
+static	void	resend_query (ResRQ *);
+static	int	proc_answer (ResRQ *, HEADER *, char *, char *);
+static	int	query_name (char *, int, int, ResRQ *);
+static	aCache	*make_cache (ResRQ *), *rem_list (aCache *);
+static	aCache	*find_cache_name (char *);
+static	aCache	*find_cache_number (ResRQ *, char *);
+static	int	add_request (ResRQ *);
+static	ResRQ	*make_request (Link *);
+static	int	send_res_msg (char *, int, int);
+static	ResRQ	*find_id (int);
+static	int	hash_number (unsigned char *);
+static	void	update_list (ResRQ *, aCache *);
+static	int	hash_name (char *);
+static	int	bad_hostname (char *, int);
 
 static	struct cacheinfo {
 	int	ca_adds;
@@ -78,8 +79,7 @@ static	struct	resinfo {
 	int	re_unkrep;
 } reinfo;
 
-int	init_resolver(op)
-int	op;
+int	init_resolver(int op)
 {
 	int	ret = 0;
 
@@ -98,10 +98,13 @@ int	op;
 		    {
 			ircd_res.nscount = 1;
 #ifdef INET6
-			/* still IPv4 */
-			ircd_res.nsaddr_list[0].sin_addr.s_addr =
-			  inet_pton(AF_INET, "127.0.0.1",
-				    &ircd_res.nsaddr_list[0].sin_addr.s_addr);
+			if (!inetpton(AF_INET6, "::1",
+				    &ircd_res.nsaddr_list[0].sin6_addr.s6_addr))
+			{
+				bcopy(minus_one,
+					ircd_res.nsaddr_list[0].sin6_addr.s6_addr,
+					IN6ADDRSZ);
+			}
 #else
 			ircd_res.nsaddr_list[0].sin_addr.s_addr =
 				inetaddr("127.0.0.1");
@@ -114,12 +117,30 @@ int	op;
 		int	on = 0;
 
 #ifdef INET6
-		/* still IPv4 */
-		ret = resfd = socket(AF_INET, SOCK_DGRAM, 0);
+		ret = resfd = socket(AF_INET6, SOCK_DGRAM, 0);
 #else
 		ret = resfd = socket(AF_INET, SOCK_DGRAM, 0);
 #endif
 		(void) SETSOCKOPT(ret, SOL_SOCKET, SO_BROADCAST, &on, on);
+
+		/* The following frame is a hack to allow resolving
+		 * in FreeBSD jail(). As it is harmless elsewhere, it is
+		 * not #ifdef-ed.
+		 * Note that currently IPv6 within jail() is not
+		 * supported by the FreeBSD.
+		 */
+		{
+			struct SOCKADDR_IN res_addr;
+
+			memset(&res_addr, 0, sizeof(res_addr));
+			res_addr.SIN_FAMILY = AFINET;
+#ifdef INET6
+			res_addr.sin6_addr = in6addr_any;
+#else
+			res_addr.sin_addr.s_addr = htonl(INADDR_ANY);
+#endif
+			bind(resfd, (SAP) &res_addr, sizeof(res_addr));
+		}
 	    }
 #ifdef DEBUG
 	if (op & RES_INITDEBG);
@@ -135,8 +156,7 @@ int	op;
 	return ret;
 }
 
-static	int	add_request(new)
-ResRQ *new;
+static	int	add_request(ResRQ *new)
 {
 	if (!new)
 		return -1;
@@ -156,8 +176,7 @@ ResRQ *new;
  * remove a request from the list. This must also free any memory that has
  * been allocated for temporary storage of DNS results.
  */
-static	void	rem_request(old)
-ResRQ	*old;
+static	void	rem_request(ResRQ *old)
 {
 	Reg	ResRQ	**rptr, *r2ptr = NULL;
 	Reg	int	i;
@@ -179,13 +198,13 @@ ResRQ	*old;
 #endif
 	r2ptr = old;
 	if (r2ptr->he.h_name)
-		MyFree((char *)r2ptr->he.h_name);
+		MyFree(r2ptr->he.h_name);
 	for (i = 0; i < MAXALIASES; i++)
 		if ((s = r2ptr->he.h_aliases[i]))
 			MyFree(s);
 	if (r2ptr->name)
 		MyFree(r2ptr->name);
-	MyFree((char *)r2ptr);
+	MyFree(r2ptr);
 
 	return;
 }
@@ -193,8 +212,7 @@ ResRQ	*old;
 /*
  * Create a DNS request record for the server.
  */
-static	ResRQ	*make_request(lp)
-Link	*lp;
+static	ResRQ	*make_request(Link *lp)
 {
 	Reg	ResRQ	*nreq;
 
@@ -221,8 +239,7 @@ Link	*lp;
  * Remove queries from the list which have been there too long without
  * being resolved.
  */
-time_t	timeout_query_list(now)
-time_t	now;
+time_t	timeout_query_list(time_t now)
 {
 	Reg	ResRQ	*rptr, *r2ptr;
 	Reg	time_t	next = 0, tout;
@@ -230,12 +247,13 @@ time_t	now;
 
 	Debug((DEBUG_DNS,"timeout_query_list at %s",myctime(now)));
 	for (rptr = first; rptr; rptr = r2ptr)
-	    {
+	{
 		r2ptr = rptr->next;
 		tout = rptr->sentat + rptr->timeout;
 		if (now >= tout)
+		{
 			if (--rptr->retries <= 0)
-			    {
+			{
 #ifdef DEBUG
 				Debug((DEBUG_ERROR,"timeout %x now %d cptr %x",
 				       rptr, now, rptr->cinfo.value.cptr));
@@ -243,7 +261,7 @@ time_t	now;
 				reinfo.re_timeouts++;
 				cptr = rptr->cinfo.value.cptr;
 				switch (rptr->cinfo.flags)
-				    {
+				{
 				case ASYNC_CLIENT :
 #if defined(USE_IAUTH)
 					sendto_iauth("%d d", cptr->fd);
@@ -255,12 +273,12 @@ time_t	now;
 						    "Host %s unknown",
 						    rptr->name);
 					break;
-				    }
+				}
 				rem_request(rptr);
 				continue;
-			    }
+			}
 			else
-			    {
+			{
 				rptr->sentat = now;
 				rptr->timeout += rptr->timeout;
 				resend_query(rptr);
@@ -270,10 +288,13 @@ time_t	now;
 				       rptr, now, rptr->retries,
 				       rptr->cinfo.value.cptr));
 #endif
-			    }
+			}
+		}
 		if (!next || tout < next)
+		{
 			next = tout;
-	    }
+		}
+	}
 	return (next > now) ? next : (now + AR_TTL);
 }
 
@@ -281,8 +302,7 @@ time_t	now;
  * del_queries - called by the server to cleanup outstanding queries for
  * which there no longer exist clients or conf lines.
  */
-void	del_queries(cp)
-char	*cp;
+void	del_queries(char *cp)
 {
 	Reg	ResRQ	*rptr, *r2ptr;
 
@@ -301,9 +321,7 @@ char	*cp;
  * isnt present. Returns number of messages successfully sent to 
  * nameservers or -1 if no successful sends.
  */
-static	int	send_res_msg(msg, len, rcount)
-char	*msg;
-int	len, rcount;
+static	int	send_res_msg(char *msg, int len, int rcount)
 {
 	Reg	int	i;
 	int	sent = 0, max;
@@ -320,15 +338,14 @@ int	len, rcount;
 	for (i = 0; i < max; i++)
 	    {
 #ifdef INET6
-		/* still IPv4 */
-		ircd_res.nsaddr_list[i].sin_family = AF_INET;
+		ircd_res.nsaddr_list[i].sin6_family = AF_INET6;
 #else
 		ircd_res.nsaddr_list[i].sin_family = AF_INET;
 #endif
 #ifdef INET6
 		if (sendto(resfd, msg, len, 0,
 			   (struct sockaddr *)&(ircd_res.nsaddr_list[i]),
-			   sizeof(struct sockaddr)) == len)
+			   sizeof(struct sockaddr_in6)) == len)
 #else
 		if (sendto(resfd, msg, len, 0,
 			   (struct sockaddr *)&(ircd_res.nsaddr_list[i]),
@@ -351,8 +368,7 @@ int	len, rcount;
 /*
  * find a dns request id (id is determined by dn_mkquery)
  */
-static	ResRQ	*find_id(id)
-int	id;
+static	ResRQ	*find_id(int id)
 {
 	Reg	ResRQ	*rptr;
 
@@ -368,10 +384,7 @@ int	id;
  * returns the host info if found in cache, or NULL when it doesn't
  * know it yet.
  */
-struct	hostent	*gethost_byname_type(name, lp, type)
-char	*name;
-Link	*lp;
-int	type;
+struct	hostent	*gethost_byname_type(char *name, Link *lp, int type)
 {
 	Reg	aCache	*cp;
 
@@ -401,9 +414,7 @@ struct	hostent	*gethost_byname(char *name, Link *lp)
 #endif
 }
 
-struct	hostent	*gethost_byaddr(addr, lp)
-char	*addr;
-Link	*lp;
+struct	hostent	*gethost_byaddr(char *addr, Link *lp)
 {
 	aCache	*cp;
 
@@ -416,11 +427,7 @@ Link	*lp;
 	return NULL;
 }
 
-static	int	do_query_name(lp, name, rptr, type)
-Link	*lp;
-char	*name;
-Reg	ResRQ	*rptr;
-int	type;
+static	int	do_query_name(Link *lp, char *name, ResRQ *rptr, int type)
 {
 	char	hname[HOSTLEN+1];
 	int	len;
@@ -459,10 +466,7 @@ int	type;
 /*
  * Use this to do reverse IP# lookups.
  */
-static	int	do_query_number(lp, numb, rptr)
-Link	*lp;
-struct	IN_ADDR	*numb;
-Reg	ResRQ	*rptr;
+static	int	do_query_number(Link *lp, struct IN_ADDR *numb, ResRQ *rptr)
 {
 	char	ipbuf[128];
 	Reg	u_char	*cp;
@@ -479,7 +483,11 @@ Reg	ResRQ	*rptr;
 	    }
 	else
 	    {
-		(void)sprintf(ipbuf, "%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.%x.ip6.%s.",
+		(void)sprintf(ipbuf,
+			"%x.%x.%x.%x.%x.%x.%x.%x."
+			"%x.%x.%x.%x.%x.%x.%x.%x."
+			"%x.%x.%x.%x.%x.%x.%x.%x."
+			"%x.%x.%x.%x.%x.%x.%x.%x.ip6.%s.",
 		(u_int)(cp[15]&0xf), (u_int)(cp[15]>>4),
 		(u_int)(cp[14]&0xf), (u_int)(cp[14]>>4),
 		(u_int)(cp[13]&0xf), (u_int)(cp[13]>>4),
@@ -497,7 +505,9 @@ Reg	ResRQ	*rptr;
 		(u_int)(cp[1]&0xf), (u_int)(cp[1]>>4),
 		(u_int)(cp[0]&0xf), (u_int)(cp[0]>>4),
 #ifdef SIXBONE_HACK
-		(cp[0] == 0x3f && cp[1] == 0xfe) ? "int" :
+		/* use ip6.arpa for 2001, ip6.int for all the rest
+		   (idea from Axu) --B. */
+		(cp[0] != 0x20 || cp[1] != 0x01) ? "int" :
 #endif
 		"arpa");
 	    }
@@ -529,10 +539,7 @@ Reg	ResRQ	*rptr;
 /*
  * generate a query based on class, type and name.
  */
-static	int	query_name(name, class, type, rptr)
-char	*name;
-int	class, type;
-ResRQ	*rptr;
+static	int	query_name(char *name, int class, int type, ResRQ *rptr)
 {
 	struct	timeval	tv;
 	char	buf[MAXPACKET];
@@ -575,8 +582,7 @@ ResRQ	*rptr;
 	return 0;
 }
 
-static	void	resend_query(rptr)
-ResRQ	*rptr;
+static	void	resend_query(ResRQ *rptr)
 {
 	if (rptr->resend == 0)
 		return;
@@ -601,10 +607,7 @@ ResRQ	*rptr;
 /*
  * process name server reply.
  */
-static	int	proc_answer(rptr, hptr, buf, eob)
-ResRQ	*rptr;
-char	*buf, *eob;
-HEADER	*hptr;
+static	int	proc_answer(ResRQ *rptr, HEADER *hptr, char *buf, char *eob)
 {
 	Reg	char	*cp, **alias;
 	Reg	struct	hent	*hp;
@@ -661,6 +664,7 @@ HEADER	*hptr;
 			}
 			if (len)
 			{
+				/* probably will never happen */
 				hostbuf[len++] = '.';
 			}
 			strcpy(hostbuf + len, ircd_res.defdname);
@@ -670,7 +674,7 @@ HEADER	*hptr;
 		/* Check that it's a possible reply to the request we send. */
 		if (rptr->type != type && type != T_CNAME)
 		{
-			sendto_flag(SCH_ERROR, "Wrong reply type looking up %s."
+			sendto_flag(SCH_ERROR, "Wrong reply type looking up %s. "
 				"Got: %d, expected %d.", hostbuf,
 				type, rptr->type);
 			cp += dlen;
@@ -726,11 +730,26 @@ HEADER	*hptr;
 			       inetntoa((char *)adr),
 			       hostbuf));
 #endif
-			if (!hp->h_name && len < HOSTLEN)
-			    {
-				hp->h_name =(char *)MyMalloc(len+1);
-				(void)strcpy(hp->h_name, hostbuf);
-			    }
+			if(len < HOSTLEN)
+			{
+				/* if we have no hostname currently,
+				 * now we have one, else we have just
+				 * another alias. -- BitKoenig */
+				if(!hp->h_name)
+				{
+					hp->h_name = (char *)MyMalloc(len+1);
+					(void)strcpy(hp->h_name, hostbuf);
+			    	}
+				else
+				/* don't keep it if we already have it */
+				if(strcasecmp(hp->h_name, hostbuf) &&
+					alias < &(hp->h_aliases[MAXALIASES-1]))
+				{
+					*alias = (char *)MyMalloc(len + 1);
+					(void)strcpy(*alias++, hostbuf);
+					*alias = NULL;
+				}
+			}
 			ans++;
 			adr++;
 			cp += dlen;
@@ -776,13 +795,23 @@ HEADER	*hptr;
 				return -1;
 			    }
 			Debug((DEBUG_INFO,"got cname %s",hostbuf));
+#if 0
+			/* This is overcautious (we do not use CNAME in any
+			   other way than follow it to get the PTR) and breaks
+			   RFC2317 (we do not allow '/' in PTRs). */
 			if (bad_hostname(hostbuf, len))
 				return -1; /* a break would be enough here */
+#endif
 			if (alias >= &(hp->h_aliases[MAXALIASES-1]))
 				break;
-			*alias = (char *)MyMalloc(len + 1);
-			(void)strcpy(*alias++, hostbuf);
-			*alias = NULL;
+			/* if we already have this alias as hostname,
+			 * don't keep it twice --BitKoenig */
+			if(hp->h_name && strcasecmp(hp->h_name, hostbuf))
+			{
+				*alias = (char *)MyMalloc(len + 1);
+				(void)strcpy(*alias++, hostbuf);
+				*alias = NULL;
+			}
 			ans++;
 			break;
 		default :
@@ -800,15 +829,14 @@ HEADER	*hptr;
 /*
  * read a dns reply from the nameserver and process it.
  */
-struct	hostent	*get_res(lp)
-char	*lp;
+struct	hostent	*get_res(char *lp)
 {
 	static	char	buf[sizeof(HEADER) + MAXPACKET];
 	Reg	HEADER	*hptr;
 	Reg	ResRQ	*rptr = NULL;
 	aCache	*cp = NULL;
 #ifdef INET6
-	struct	sockaddr_in	sin;
+	struct	sockaddr_in6	sin;
 #else
 	struct	sockaddr_in	sin;
 #endif
@@ -856,10 +884,10 @@ char	*lp;
 
 	for (a = 0; a < max; a++)
 #ifdef INET6
-		if (!ircd_res.nsaddr_list[a].sin_addr.s_addr ||
-		    !bcmp((char *)&sin.sin_addr,
-			  (char *)&ircd_res.nsaddr_list[a].sin_addr,
-			  sizeof(struct in_addr)))
+		if (!ircd_res.nsaddr_list[a].SIN_ADDR.S_ADDR ||
+		    !bcmp((char *)&sin.SIN_ADDR,
+			  (char *)&ircd_res.nsaddr_list[a].SIN_ADDR,
+			  sizeof(struct IN_ADDR)))
 #else
 		if (!ircd_res.nsaddr_list[a].sin_addr.s_addr ||
 		    !bcmp((char *)&sin.sin_addr,
@@ -909,17 +937,14 @@ char	*lp;
 	    }
 	a = proc_answer(rptr, hptr, buf, buf+rc);
 	if (a == -1) {
+		sprintf(buffer, "Bad hostname returned for %s",
 #ifdef	INET6
-		SPRINTF(buffer, "Bad hostname returned from %s for %s",
-			inetntop(AF_INET, &sin.sin_addr, mydummy2, 
-				MYDUMMY_SIZE),
 			inetntop(AF_INET6, rptr->he.h_addr.s6_addr,
-				mydummy, MYDUMMY_SIZE));
+				mydummy, MYDUMMY_SIZE)
 #else
-		SPRINTF(buffer, "Bad hostname returned from %s for ", 
-			inetntoa((char *)&sin.sin_addr));
-		strcat(buffer, inetntoa((char *)&rptr->he.h_addr));
+			inetntoa((char *)&rptr->he.h_addr)
 #endif
+		);
 		sendto_flag(SCH_ERROR, "%s", buffer);
 		Debug((DEBUG_DNS, "%s", buffer));
 	}
@@ -944,7 +969,7 @@ char	*lp;
 			rptr->he.h_name, inetntoa((char *)&rptr->he.h_addr)));
 #endif
 		/*
-		 * Lookup the 'authoritive' name that we were given for the
+		 * Lookup the 'authoritative' name that we were given for the
 		 * ip#.  By using this call rather than regenerating the
 		 * type we automatically gain the use of the cache with no
 		 * extra kludges.
@@ -969,8 +994,12 @@ char	*lp;
 		 * If name wasn't found, a request has been queued and it will
 		 * be the last one queued.  This is rather nasty way to keep
 		 * a host alias with the query. -avalon
+		 * We also need 'authoritative' name to be kept --BitKoenig
 		 */
-		if (!hp2 && rptr->he.h_aliases[0])
+		if (!hp2)
+		{
+			last->he.h_name = rptr->he.h_name;
+			rptr->he.h_name = NULL;
 			for (a = 0; rptr->he.h_aliases[a]; a++)
 			    {
 				Debug((DEBUG_DNS, "Copied CNAME %s for %s",
@@ -979,6 +1008,7 @@ char	*lp;
 				last->he.h_aliases[a] = rptr->he.h_aliases[a];
 				rptr->he.h_aliases[a] = NULL;
 			    }
+		}
 
 		rem_request(rptr);
 		return hp2;
@@ -1007,7 +1037,7 @@ getres_err:
 	if (rptr)
 	    {
 		if (h_errno != TRY_AGAIN)
-		    {
+		{
 			/*
 			 * If we havent tried with the default domain and its
 			 * set, then give it a try next.
@@ -1025,21 +1055,20 @@ getres_err:
 			{
 				rptr->type = T_A;
 				query_name(rptr->name, C_IN, T_A, rptr);
-				Debug((DEBUG_DNS, "getres_err: didn't work "
+				Debug((DEBUG_DNS,"getres_err: didn't work "
 					"with T_AAAA, now also trying with "
 					"T_A for %s", rptr->name));
 			}
 #endif
 			resend_query(rptr);
-		    }
+		}
 		else if (lp)
 			bcopy((char *)&rptr->cinfo, lp, sizeof(Link));
 	    }
 	return (struct hostent *)NULL;
 }
 
-static	int	hash_number(ip)
-Reg	u_char	*ip;
+static	int	hash_number(u_char *ip)
 {
 	Reg	u_int	hashv = 0;
 
@@ -1066,8 +1095,7 @@ Reg	u_char	*ip;
 	return (hashv);
 }
 
-static	int	hash_name(name)
-register	char	*name;
+static	int	hash_name(char *name)
 {
 	Reg	u_int	hashv = 0;
 
@@ -1080,8 +1108,7 @@ register	char	*name;
 /*
 ** Add a new cache item to the queue and hash table.
 */
-static	aCache	*add_to_cache(ocp)
-Reg	aCache	*ocp;
+static	aCache	*add_to_cache(aCache *ocp)
 {
 	Reg	aCache	*cp = NULL;
 	Reg	int	hashv;
@@ -1137,9 +1164,7 @@ Reg	aCache	*ocp;
 ** it already contains the correct expire time, if it is a new entry. Old
 ** entries have the expirey time updated.
 */
-static	void	update_list(rptr, cachep)
-ResRQ	*rptr;
-aCache	*cachep;
+static	void	update_list(ResRQ *rptr, aCache *cachep)
 {
 	Reg	aCache	**cpp, *cp = cachep;
 	Reg	char	*s, *t, **base;
@@ -1267,8 +1292,7 @@ aCache	*cachep;
 	return;
 }
 
-static	aCache	*find_cache_name(name)
-char	*name;
+static	aCache	*find_cache_name(char *name)
 {
 	Reg	aCache	*cp;
 	Reg	char	*s;
@@ -1313,9 +1337,7 @@ char	*name;
 /*
  * find a cache entry by ip# and update its expire time
  */
-static	aCache	*find_cache_number(rptr, numb)
-ResRQ	*rptr;
-char	*numb;
+static	aCache	*find_cache_number(ResRQ *rptr, char *numb)
 {
 	Reg	aCache	*cp;
 	Reg	int	hashv,i;
@@ -1389,8 +1411,7 @@ char	*numb;
 	return NULL;
 }
 
-static	aCache	*make_cache(rptr)
-ResRQ	*rptr;
+static	aCache	*make_cache(ResRQ *rptr)
 {
 	Reg	aCache	*cp;
 	Reg	int	i, n;
@@ -1479,8 +1500,7 @@ ResRQ	*rptr;
 /*
  * rem_list
  */
-static	aCache	*rem_list(cp)
-aCache	*cp;
+static	aCache	*rem_list(aCache *cp)
 {
 	aCache	**cpp, *cr = cp->list_next;
 
@@ -1491,7 +1511,7 @@ aCache	*cp;
 		if (*cpp == cp)
 		    {
 			*cpp = cp->list_next;
-			MyFree((char *)cp);
+			MyFree(cp);
 			break;
 		    }
 	return cr;
@@ -1503,8 +1523,7 @@ aCache	*cp;
  *     delete a cache entry from the cache structures and lists and return
  *     all memory used for the cache back to the memory pool.
  */
-static	void	rem_cache(ocp)
-aCache	*ocp;
+static	void	rem_cache(aCache *ocp)
 {
 	Reg	aCache	**cp;
 	Reg	struct	hostent *hp = &ocp->he;
@@ -1578,7 +1597,7 @@ aCache	*ocp;
 	    {
 		for (hashv = 0; hp->h_aliases[hashv]; hashv++)
 			MyFree(hp->h_aliases[hashv]);
-		MyFree((char *)hp->h_aliases);
+		MyFree(hp->h_aliases);
 	    }
 
 	/*
@@ -1587,11 +1606,11 @@ aCache	*ocp;
 	if (hp->h_addr_list)
 	    {
 		if (*hp->h_addr_list)
-			MyFree((char *)*hp->h_addr_list);
-		MyFree((char *)hp->h_addr_list);
+			MyFree(*hp->h_addr_list);
+		MyFree(hp->h_addr_list);
 	    }
 
-	MyFree((char *)ocp);
+	MyFree(ocp);
 
 	incache--;
 	cainfo.ca_dels++;
@@ -1603,8 +1622,7 @@ aCache	*ocp;
  * removes entries from the cache which are older than their expirey times.
  * returns the time at which the server should next poll the cache.
  */
-time_t	expire_cache(now)
-time_t	now;
+time_t	expire_cache(time_t now)
 {
 	Reg	aCache	*cp, *cp2;
 	Reg	time_t	next = 0;
@@ -1635,14 +1653,14 @@ void	flush_cache()
 		rem_cache(cp);
 }
 
-int	m_dns(cptr, sptr, parc, parv)
-aClient *cptr, *sptr;
-int	parc;
-char	*parv[];
+int	m_dns(aClient *cptr, aClient *sptr, int parc, char *parv[])
 {
 	Reg	aCache	*cp;
 	Reg	int	i;
 
+	if (!is_allowed(sptr, ACL_DNS))
+		return m_nopriv(cptr, sptr, parc, parv);
+
 	if (parv[1] && *parv[1] == 'l') {
 		for(cp = cachetop; cp; cp = cp->list_next)
 		    {
@@ -1693,9 +1711,7 @@ char	*parv[];
 	return 2;
 }
 
-u_long	cres_mem(sptr, nick)
-aClient	*sptr;
-char	*nick;
+u_long	cres_mem(aClient *sptr, char *nick)
 {
 	register aCache	*c = cachetop;
 	register struct	hostent	*h;
@@ -1735,9 +1751,7 @@ char	*nick;
 }
 
 
-static	int	bad_hostname(name, len)
-char *name;
-int len;
+static	int	bad_hostname(char *name, int len)
 {
 	char	*s, c;
 
@@ -1745,7 +1759,7 @@ int len;
 #ifdef RESTRICT_HOSTNAMES
 	{
 		/* basic character set */
-		if (isascii(c) && isalnum(c))
+		if (isalnum(c))
 			continue;
 		
 		/* special case: hyphen */
@@ -1777,3 +1791,4 @@ int len;
 #endif /* RESTRICT_HOSTNAMES */
 	return 0;
 }
+
