@@ -19,7 +19,7 @@
  */
 
 #ifndef lint
-static  char rcsid[] = "@(#)$Id: s_debug.c,v 1.28.2.4 2001/10/19 18:44:27 chopin Exp $";
+static const volatile char rcsid[] = "@(#)$Id: s_debug.c,v 1.47 2004/10/01 20:22:15 chopin Exp $";
 #endif
 
 #include "os.h"
@@ -48,9 +48,6 @@ char	serveropts[] = {
 #ifdef	RANDOM_NDELAY
 'd',
 #endif
-#if defined(LOCOP_REHASH) && defined(OPER_REHASH)
-'e',
-#endif
 #ifdef	OPER_REHASH
 'E',
 #endif
@@ -60,34 +57,28 @@ char	serveropts[] = {
 #ifdef	CLONE_CHECK
 'F',
 #endif
-#ifdef	SUN_GSO_BUG
-'g',
-#endif
 #ifdef	HUB
 'H',
 #endif
 #ifdef	BETTER_CDELAY
 'h',
 #endif
-#ifdef	SHOW_INVISIBLE_LUSERS
-'i',
-#endif
 #ifndef	NO_DEFAULT_INVISIBLE
 'I',
 #endif
-#if defined(LOCOP_DIE) && defined(OPER_DIE)
-'j',
-#endif
 #ifdef	OPER_DIE
 'J',
 #endif
 #ifdef	OPER_KILL
-# ifdef  LOCAL_KILL_ONLY
+# ifndef  OPER_KILL_REMOTE
 'k',
 # else
 'K',
 # endif
 #endif
+#ifdef FAILED_OPERLOG
+'l',
+#endif
 #ifdef	LEAST_IDLE
 'L',
 #endif
@@ -97,9 +88,6 @@ char	serveropts[] = {
 #ifdef	IDLE_FROM_MSG
 'M',
 #endif
-#ifdef	NPATH /* gone */
-'N',
-#endif
 #ifdef	BETTER_NDELAY
 'n',
 #endif
@@ -109,9 +97,6 @@ char	serveropts[] = {
 #ifdef	CRYPT_LINK_PASSWORD
 'P',
 #endif
-#if defined(LOCOP_RESTART) && defined(OPER_RESTART)
-'r',
-#endif
 #ifdef	OPER_RESTART
 'R',
 #endif
@@ -121,8 +106,8 @@ char	serveropts[] = {
 #ifdef	ENABLE_SUMMON
 'S',
 #endif
-#ifdef	OPER_REMOTE
-'t',
+#ifdef TOPIC_WHO_TIME
+'T',
 #endif
 #ifndef	NO_PREFIX
 'u',
@@ -130,12 +115,6 @@ char	serveropts[] = {
 #ifdef	ENABLE_USERS
 'U',
 #endif
-#ifdef	VALLOC
-'V',
-#endif
-#ifdef	NOWRITEALARM
-'w',
-#endif
 #ifdef	UNIXPORT
 'X',
 #endif
@@ -153,22 +132,13 @@ char	serveropts[] = {
 #ifdef DEBUGMODE
 static	char	debugbuf[2*READBUF_SIZE]; /* needs to be big.. */
 
-#if ! USE_STDARG
-void	debug(level, form, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10)
-int	level;
-char	*form, *p1, *p2, *p3, *p4, *p5, *p6, *p7, *p8, *p9, *p10;
-#else
 void	debug(int level, char *form, ...)
-#endif
 {
 	int	err = errno;
 
 #ifdef	USE_SYSLOG
 	if (level == DEBUG_ERROR)
 	    {
-#if ! USE_STDARG
-		syslog(LOG_ERR, form, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10);
-#else
 # if HAVE_VSYSLOG
 		va_list va;
 		va_start(va, form);
@@ -181,25 +151,14 @@ void	debug(int level, char *form, ...)
 		va_end(va);
 		syslog(LOG_ERR, "%s", debugbuf);
 # endif
-#endif
 	    }
 #endif
 	if ((debuglevel >= 0) && (level <= debuglevel))
 	    {
-#if ! USE_STDARG
-		(void)sprintf(debugbuf, form,
-				p1, p2, p3, p4, p5, p6, p7, p8, p9, p10);
-#else
 		va_list va;
 		va_start(va, form);
 		(void)vsprintf(debugbuf, form, va);
 		va_end(va);
-#endif
-		if (local[2])
-		    {
-			local[2]->sendM++;
-			local[2]->sendB += strlen(debugbuf);
-		    }
 		(void)fprintf(stderr, "%s\n", debugbuf);
 	    }
 	errno = err;
@@ -213,9 +172,7 @@ void	debug(int level, char *form, ...)
  * different field names for "struct rusage".
  * -avalon
  */
-void	send_usage(cptr, nick)
-aClient *cptr;
-char	*nick;
+void	send_usage(aClient *cptr, char *nick)
 {
 #if HAVE_GETRUSAGE
 	struct	rusage	rus;
@@ -318,9 +275,7 @@ char	*nick;
 	return;
 }
 
-void	send_defines(cptr, nick)
-aClient *cptr;
-char	*nick;
+void	send_defines(aClient *cptr, char *nick)
 {
     	sendto_one(cptr, ":%s %d %s :HUB:%s MS:%d", 
 		   ME, RPL_STATSDEFINE, nick,
@@ -339,42 +294,51 @@ char	*nick;
 		   ME, RPL_STATSDEFINE, nick, KILLCHASETIMELIMIT,
 		   DELAYCHASETIMELIMIT, LDELAYCHASETIMELIMIT,
 		   CLIENT_FLOOD, MAXCHANNELSPERUSER);
-	sendto_one(cptr, ":%s %d %s :H:%d N:%d U:%d R:%d T:%d C:%d P:%d K:%d",
-		   ME, RPL_STATSDEFINE, nick, HOSTLEN, NICKLEN, USERLEN,
+	sendto_one(cptr, ":%s %d %s :H:%d N:%d D:%d U:%d R:%d T:%d C:%d P:%d K:%d",
+		   ME, RPL_STATSDEFINE, nick, HOSTLEN, ONICKLEN, UIDLEN, USERLEN,
 		   REALLEN, TOPICLEN, CHANNELLEN, PASSWDLEN, KEYLEN);
 	sendto_one(cptr, ":%s %d %s :BS:%d MXR:%d MXB:%d MXBL:%d PY:%d",
 		   ME, RPL_STATSDEFINE, nick, BUFSIZE, MAXRECIPIENTS, MAXBANS,
 		   MAXBANLENGTH, MAXPENALTY);
-	sendto_one(cptr, ":%s %d %s :ZL:%d CM:%d CP:%d",
-		   ME, RPL_STATSDEFINE, nick,
+	sendto_one(cptr, ":%s %d %s :ZL:%d CM:%d CP:%d DC:%d",
+		ME, RPL_STATSDEFINE, nick,
 #ifdef	ZIP_LINKS
-		   ZIP_LEVEL,
+		ZIP_LEVEL,
 #else
-		   -1,
+		-1,
 #endif
 #ifdef	CLONE_CHECK
-		   CLONE_MAX, CLONE_PERIOD
+		CLONE_MAX, CLONE_PERIOD,
 #else
-		   -1, -1
+		-1, -1,
+#endif
+#ifdef	DELAY_CLOSE
+		DELAY_CLOSE
+#else
+		-1
+#endif
+		);
+	sendto_one(cptr, ":%s %d %s :AC:%d CA:%d S:%d SS:%d SU:%d", ME, RPL_STATSDEFINE,
+		   nick, iconf.aconnect, iconf.caccept, iconf.split,
+		   iconf.split_minservers, iconf.split_minusers);
+#ifdef CLIENTS_CHANNEL
+	sendto_one(cptr, ":%s %d %s :CCL:0x%X", ME, RPL_STATSDEFINE, nick,
+		CLIENTS_CHANNEL_LEVEL);
 #endif
-		   );
 }
 
-void	count_memory(cptr, nick, debug)
-aClient	*cptr;
-char	*nick;
-int	debug;
+void	count_memory(aClient *cptr, char *nick, int debug)
 {
 	extern	aChannel	*channel;
-	extern	aClass	*classes;
+	extern	aClass		*classes;
 	extern	aConfItem	*conf;
-	extern	int	_HASHSIZE, _CHANNELHASHSIZE;
+	extern	int		_HASHSIZE, _CHANNELHASHSIZE;
 
-	Reg	aClient	*acptr;
-	Reg	Link	*link;
-	Reg	aChannel *chptr;
-	Reg	aConfItem *aconf;
-	Reg	aClass	*cltmp;
+	Reg	aClient		*acptr;
+	Reg	Link		*link;
+	Reg	aChannel	*chptr;
+	Reg	aConfItem	*aconf;
+	Reg	aClass		*cltmp;
 
 	int	lc = 0, d_lc = 0,	/* local clients */
 		ch = 0, d_ch = 0,	/* channels */
@@ -438,30 +402,40 @@ int	debug;
 	awm = istat.is_awaymem;
 
 	if (debug)
+	{
 		for (acptr = client; acptr; acptr = acptr->next)
-		    {
+		{
 			if (MyConnect(acptr))
-			    {
+			{
 				d_lc++;
-				for (link =acptr->confs; link; link=link->next)
+				for (link = acptr->confs; link; link=link->next)
+				{
 					d_lcc++;
+				}
 			}
-			    else
+			else
+			{
 				d_rc++;
+			}
 			if (acptr->user)
-			    {
+			{
+				invLink	*ilink;
+
 				d_us++;
-				for (link = acptr->user->invited; link;
-				     link = link->next)
+				for (ilink = acptr->user->invited; ilink;
+				     ilink = ilink->next)
+				{
 					d_usi++;
+				}
 				d_usc += acptr->user->joined;
 				if (acptr->user->away)
-				    {
+				{
 					d_aw++;
 					d_awm += (strlen(acptr->user->away)+1);
-				    }
-			    }
-		    }
+				}
+			}
+		}
+	}
 
 	lcm = lc * CLIENT_LOCAL_SIZE;
 	rcm = rc * CLIENT_REMOTE_SIZE;
@@ -479,51 +453,57 @@ int	debug;
 	chu = istat.is_chanusers;
 
 	if (debug)
-	    {
+	{
 		for (chptr = channel; chptr; chptr = chptr->nextch)
-		    {
+		{
 			if (chptr->users == 0)
-			    {
+			{
 				d_chh++;
 				d_chhm+=strlen(chptr->chname)+sizeof(aChannel);
-			    }
+			}
 			else
-			    {
+			{
 				d_ch++;
 				d_chm += (strlen(chptr->chname) +
 					  sizeof(aChannel));
-			    }
+			}
 			for (link = chptr->members; link; link = link->next)
+			{
 				d_chu++;
+			}
 			for (link = chptr->invites; link; link = link->next)
+			{
 				d_chi++;
+			}
 			for (link = chptr->mlist; link; link = link->next)
-			    {
+			{
 				d_chb++;
 				d_chbm += strlen(link->value.cp) + 1;
-			    }
-		    }
+			}
+		}
 		d_chbm += d_chb * sizeof(Link);
-	    }
+	}
 
 	co = istat.is_conf;
 	com = istat.is_confmem;
 	cl = istat.is_class;
 
 	if (debug)
-	    {
+	{
 		for (aconf = conf; aconf; aconf = aconf->next)
-		    {
+		{
 			d_co++;
 			d_com += aconf->host ? strlen(aconf->host)+1 : 0;
 			d_com += aconf->passwd ? strlen(aconf->passwd)+1 : 0;
 			d_com += aconf->name ? strlen(aconf->name)+1 : 0;
 			d_com += aconf->ping ? sizeof(*aconf->ping) : 0;
 			d_com += sizeof(aConfItem);
-		    }
+		}
 		for (cltmp = classes; cltmp; cltmp = cltmp->next)
+		{
 			d_cl++;
-	    }
+		}
+	}
 
 	if (debug)
 		sendto_one(cptr, ":%s %d %s :Request processed in %u seconds",
@@ -535,34 +515,42 @@ int	debug;
 		   istat.is_auth, istat.is_authmem);
 	if (debug
 	    && (lc != d_lc || lcm != d_lcm || rc != d_rc || rcm != d_rcm))
+	{
 		sendto_one(cptr,
 			":%s %d %s :Client Local %d(%lu) Remote %d(%lu) [REAL]",
-			   me.name, RPL_STATSDEBUG, nick, d_lc, d_lcm, d_rc,
-			   d_rcm);
+			me.name, RPL_STATSDEBUG, nick, d_lc, d_lcm, d_rc,
+			d_rcm);
+	}
 	sendto_one(cptr,
 		   ":%s %d %s :Users %d in/visible %d/%d(%lu) Invites %d(%lu)",
 		   me.name, RPL_STATSDEBUG, nick, us, istat.is_user[1],
 		   istat.is_user[0], us*sizeof(anUser), usi,
 		   usi*sizeof(Link));
 	if (debug && (us != d_us || usi != d_usi))
+	{
 		sendto_one(cptr,
 			   ":%s %d %s :Users %d(%lu) Invites %d(%lu) [REAL]",
 			   me.name, RPL_STATSDEBUG, nick, d_us,
 			   d_us*sizeof(anUser), d_usi, d_usi * sizeof(Link));
+	}
 	sendto_one(cptr, ":%s %d %s :User channels %d(%lu) Aways %d(%lu)",
 		   me.name, RPL_STATSDEBUG, nick, usc, usc*sizeof(Link),
 		   aw, awm);
 	if (debug && (usc != d_usc || aw != d_aw || awm != d_awm))
+	{
 		sendto_one(cptr,
 			":%s %d %s :User channels %d(%lu) Aways %d(%lu) [REAL]",
 			   me.name, RPL_STATSDEBUG, nick, d_usc,
 			   d_usc*sizeof(Link), d_aw, d_awm);
+	}
 	sendto_one(cptr, ":%s %d %s :Attached confs %d(%lu)",
 		   me.name, RPL_STATSDEBUG, nick, lcc, lcc*sizeof(Link));
 	if (debug && lcc != d_lcc)
+	{
 		sendto_one(cptr, ":%s %d %s :Attached confs %d(%lu) [REAL]",
 			   me.name, RPL_STATSDEBUG, nick, d_lcc,
 			   d_lcc*sizeof(Link));
+	}
 
 	totcl = lcm + rcm + us*sizeof(anUser) + usc*sizeof(Link) + awm;
 	totcl += lcc*sizeof(Link) + usi*sizeof(Link);
@@ -572,15 +560,19 @@ int	debug;
 	sendto_one(cptr, ":%s %d %s :Conflines %d(%lu)",
 		   me.name, RPL_STATSDEBUG, nick, co, com);
 	if (debug && (co != d_co || com != d_com))
+	{
 		sendto_one(cptr, ":%s %d %s :Conflines %d(%lu) [REAL]",
 			   me.name, RPL_STATSDEBUG, nick, d_co, d_com);
+	}
 
 	sendto_one(cptr, ":%s %d %s :Classes %d(%lu)",
 		   me.name, RPL_STATSDEBUG, nick, cl, cl*sizeof(aClass));
 	if (debug && cl != d_cl)
+	{
 		sendto_one(cptr, ":%s %d %s :Classes %d(%lu) [REAL]",
 			   me.name, RPL_STATSDEBUG, nick, d_cl,
 			   d_cl*sizeof(aClass));
+	}
 
 	sendto_one(cptr,
    ":%s %d %s :Channels %d(%lu) Modes %d(%lu) History %d(%lu) Cache %d(%lu)",
@@ -588,19 +580,22 @@ int	debug;
 		   chhm, istat.is_cchan, istat.is_cchanmem);
 	if (debug && (ch != d_ch || chm != d_chm || chb != d_chb
 		      || chbm != d_chbm || chh != d_chh || chhm != d_chhm))
+	{
 		sendto_one(cptr,
 	       ":%s %d %s :Channels %d(%lu) Modes %d(%lu) History %d(%lu) [REAL]",
 			   me.name, RPL_STATSDEBUG, nick, d_ch, d_chm, d_chb,
 			   d_chbm, d_chh, d_chhm);
+	}
 	sendto_one(cptr, ":%s %d %s :Channel members %d(%lu) invite %d(%lu)",
 		   me.name, RPL_STATSDEBUG, nick, chu, chu*sizeof(Link),
 		   chi, chi*sizeof(Link));
 	if (debug && (chu != d_chu || chi != d_chi))
+	{
 		sendto_one(cptr,
-		   ":%s %d %s :Channel members %d(%lu) invite %d(%lu) [REAL]",
-		   me.name, RPL_STATSDEBUG, nick, d_chu, d_chu*sizeof(Link),
-		   d_chi, d_chi*sizeof(Link));
-
+			":%s %d %s :Channel members %d(%lu) invite %d(%lu) "
+			"[REAL]", me.name, RPL_STATSDEBUG, nick, d_chu, 
+			d_chu*sizeof(Link), d_chi, d_chi*sizeof(Link));
+	}
 	totch = chm + chhm + chbm + chu*sizeof(Link) + chi*sizeof(Link);
 	d_totch = d_chm + d_chhm + d_chbm + d_chu*sizeof(Link)
 		  + d_chi*sizeof(Link);
@@ -611,33 +606,38 @@ int	debug;
 		   wwa, wwam, wwuw, wwuw*sizeof(Link));
 	if (debug && (wwu != d_wwu || wwa != d_wwa || wwam != d_wwam
 		      || wwuw != d_wwuw))
+	{
 		sendto_one(cptr,
-	     ":%s %d %s :Whowas users %d(%lu) away %d(%lu) links %d(%lu) [REAL]",
-		   me.name, RPL_STATSDEBUG, nick, d_wwu, d_wwu*sizeof(anUser),
-		   d_wwa, d_wwam, d_wwuw, d_wwuw*sizeof(Link));
+			":%s %d %s :Whowas users %d(%lu) away %d(%lu) "
+			"links %d(%lu) [REAL]", me.name, RPL_STATSDEBUG,
+			nick, d_wwu, d_wwu*sizeof(anUser),
+			d_wwa, d_wwam, d_wwuw, d_wwuw*sizeof(Link));
+	}
 	sendto_one(cptr, ":%s %d %s :Whowas array %d(%lu) Delay array %d(%lu)",
 		   me.name, RPL_STATSDEBUG, nick, ww_size, wwm, lk_size, dm);
 	if (debug && (wwm != d_wwm || dm != d_dm))
+	{
 		sendto_one(cptr,
-		   ":%s %d %s :Whowas array %d(%lu) Delay array %d(%lu) [REAL]",
-		   me.name, RPL_STATSDEBUG, nick, ww_size, d_wwm, lk_size,
-		   d_dm);
+			":%s %d %s :Whowas array %d(%lu) Delay array %d(%lu) "
+			"[REAL]", me.name, RPL_STATSDEBUG, nick, ww_size,
+			d_wwm, lk_size, d_dm);
+	}
 
 	totww = wwu*sizeof(anUser) + wwam + wwm;
 	d_totww = d_wwu*sizeof(anUser) + d_wwam + d_wwm;
 
 	sendto_one(cptr, ":%s %d %s :Hash: client %d(%lu) chan %d(%lu)",
-		   me.name, RPL_STATSDEBUG, nick, _HASHSIZE,
-		   sizeof(aHashEntry) * _HASHSIZE,
-		   _CHANNELHASHSIZE, sizeof(aHashEntry) * _CHANNELHASHSIZE);
+		me.name, RPL_STATSDEBUG, nick, _HASHSIZE,
+		sizeof(aHashEntry) * _HASHSIZE,
+		_CHANNELHASHSIZE, sizeof(aHashEntry) * _CHANNELHASHSIZE);
 	d_db = db = istat.is_dbufnow * sizeof(dbufbuf);
 	db = istat.is_dbufnow * sizeof(dbufbuf);
 	sendto_one(cptr,
-		   ":%s %d %s :Dbuf blocks %lu(%lu) (> %lu [%lu]) (%lu < %lu) [%lu]",
-		   me.name, RPL_STATSDEBUG, nick, istat.is_dbufnow, db,
-		   istat.is_dbuf,
-		   (u_int) (((u_int)BUFFERPOOL) / ((u_int)sizeof(dbufbuf))),
-		   istat.is_dbufuse, istat.is_dbufmax, istat.is_dbufmore);
+		":%s %d %s :Dbuf blocks %lu(%lu) (> %lu [%lu]) (%lu < %lu) "
+		"[%lu]", me.name, RPL_STATSDEBUG, nick, istat.is_dbufnow, db,
+		istat.is_dbuf,
+		(u_int) (((u_int)BUFFERPOOL) / ((u_int)sizeof(dbufbuf))),
+		istat.is_dbufuse, istat.is_dbufmax, istat.is_dbufmore);
 
 	d_rm = rm = cres_mem(cptr, nick);
 
@@ -652,16 +652,17 @@ int	debug;
 	sendto_one(cptr, ":%s %d %s :Total: ww %lu ch %lu cl %lu co %lu db %lu",
 		   me.name, RPL_STATSDEBUG, nick, totww, totch, totcl, com,db);
 	if (debug && tot != d_tot)
-	    {
+	{
 		sendto_one(cptr,
-		   ":%s %d %s :Total: ww %lu ch %lu cl %lu co %lu db %lu [REAL]",
-		   me.name, RPL_STATSDEBUG, nick, d_totww, d_totch, d_totcl,
-		   d_com, d_db);
+			":%s %d %s :Total: ww %lu ch %lu cl %lu co %lu "
+			"db %lu [REAL]", me.name, RPL_STATSDEBUG, nick,
+			d_totww, d_totch, d_totcl, d_com, d_db);
 		sendto_one(cptr, ":%s %d %s :TOTAL: %lu [REAL]",
-			   me.name, RPL_STATSDEBUG, nick, d_tot);
-	    }
+			me.name, RPL_STATSDEBUG, nick, d_tot);
+	}
 	sendto_one(cptr, ":%s %d %s :TOTAL: %d sbrk(0)-etext: %u",
 		   me.name, RPL_STATSDEBUG, nick, tot,
 		   (u_long)sbrk((size_t)0)-(u_long)sbrk0);
 	return;
 }
+
